/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

static int getch(fawk_ctx_t *ctx)
{
	int ret;
	if (ctx->parser.pushback > 0) {
		ret = ctx->parser.pushback;
		ctx->parser.pushback = -1;
	}
	else
		ret = ctx->parser.get_char(ctx, ctx->parser.isp);
	if (ret == EOF) ctx->parser.in_eof = 1;
	else if (ret == '\n') { ctx->parser.isp->line++; ctx->parser.isp->last_col = ctx->parser.isp->col; ctx->parser.isp->col = 0; }
	else { ctx->parser.isp->col++; }
	return ret;
}

static void ungetch(fawk_ctx_t *ctx, int chr)
{
	fawk_assert(ctx->parser.pushback <= 0);
	ctx->parser.pushback = chr;
	if (chr == '\n') { ctx->parser.isp->line--; ctx->parser.isp->col = ctx->parser.isp->last_col; }
	ctx->parser.isp->col--;
}

#define append(chr, bad) \
do { \
	if (ctx->parser.used >= ctx->parser.alloced) { \
		char *bf; \
		ctx->parser.alloced += 256; \
		if ((bf = fawk_realloc(ctx, ctx->parser.buff, ctx->parser.alloced)) == NULL) { ctx->parser.alloced = 0; bad; }  \
		ctx->parser.buff = bf; \
	} \
	ctx->parser.buff[ctx->parser.used++] = chr; \
} while(0)

#define del_last() ctx->parser.used--

static void fawk_readup(fawk_ctx_t *ctx, const char *accept)
{
	int c;
	do {
		c = getch(ctx);
		append(c, return);
	} while((c != EOF) && (strchr(accept, c) != NULL));
	ungetch(ctx, c);
	del_last();
}

static void readtil(fawk_ctx_t *ctx, const char *reject)
{
	int c;
	do {
		c = getch(ctx);
		append(c, return);
	} while((c != EOF) && (strchr(reject, c) == NULL));
	ungetch(ctx, c);
	del_last();
}

static int read_strlit(fawk_ctx_t *ctx, char term)
{
	int chr, len;
	for(len = 0;;len++) {
		chr = getch(ctx);
		switch(chr) {
			case '\\':
				chr = getch(ctx);
				switch(chr) {
					case 'n': append('\n', return 0); break;
					case 't': append('\t', return 0); break;
					case '0': append('\0', return 0); break;
					default:  append(chr, return 0);
				}
				break;
			default: if ((chr == term) || (chr == EOF)) { append('\0', return 0); return len; }
			append(chr, return 0);
		}
	}
}

static int read_numeric(fawk_ctx_t *ctx, fawk_num_t *dst, int had_decimal_dot, int numtoken)
{
	int chr, chr2, had_e = 0;

	next:;
	chr = getch(ctx); append(chr, return -1);
	if (isdigit(chr)) goto next;
	if ((chr == '.') && (!had_decimal_dot)) { had_decimal_dot = 1; goto next; }
	if (((chr == 'e') || (chr == 'E')) && (!had_e)) { /* 1.2e-5 */
		had_e = 1;
		chr = getch(ctx); append(chr, return -1);
		if (isdigit(chr)) goto next;
		if ((chr == '+') || (chr == '-')) {
			chr2 = getch(ctx);
			if (isdigit(chr2)) { append(chr2, return -1); goto next; }
			LIBFAWK_ERROR(ctx, "invalid numeric: e+ or e- must be followed by a digit", ctx->parser.isp->fn, ctx->parser.isp->line+1, ctx->parser.isp->col+1,  -1);
		}
		LIBFAWK_ERROR(ctx, "invalid numeric: e must be followed by sign or digit", ctx->parser.isp->fn, ctx->parser.isp->line+1, ctx->parser.isp->col+1,  -1);
	}
	ungetch(ctx, chr); del_last();
	append('\0', return -1);
	*dst = fawk_strtod(ctx->parser.buff);

	return numtoken;
}

/* Generic rule for tokens like '*=' */
#define case_op_ch2(opchr, chr2, eqtok) \
		case opchr: \
			nchr = getch(ctx); append(chr, return -1); \
			if (nchr == chr2) return eqtok; \
			else { ungetch(ctx, nchr); return chr; } \
			break

#define lex_include(restart_stmt) \
	if (ctx->parser.include != NULL) { \
		char *fn = ctx->parser.buff; \
		fawk_readup(ctx, " \t"); \
		readtil(ctx, "\""); chr = getch(ctx); \
		if (chr == '\"') { ctx->parser.used = 0; readtil(ctx, "\""); append('\0', return -1); getch(ctx); } \
		ctx->parser.isp++; \
		if (ctx->parser.isp == &ctx->parser.include_stack[FAWK_MAX_INCLUDE_STACK]) { \
			sprintf(tmp, "Includes nested too deep\n"); \
			LIBFAWK_ERROR(ctx, tmp, ctx->parser.isp->fn, ctx->parser.isp->line+1, ctx->parser.isp->col+1,  -1); \
		} \
		ctx->parser.isp->fn = fawk_strdup(ctx, fn); if (ctx->parser.isp->fn == NULL) { return -1; } \
		ctx->parser.isp->line = ctx->parser.isp->col = 0; \
		if (ctx->parser.include(ctx, ctx->parser.isp, 1, &ctx->parser.isp[-1]) != 0) { \
			fawk_free(ctx, ctx->parser.isp->fn); ctx->parser.isp->fn = NULL; \
			return -1; \
		} \
		{restart_stmt;} \
	} \
	sprintf(tmp, "Include not supported by the host application\n"); \
	LIBFAWK_ERROR(ctx, tmp, ctx->parser.isp->fn, ctx->parser.isp->line+1, ctx->parser.isp->col+1,   -1);

#define lex_got_eof(restart_stmt) \
	handle_eof:; ctx->parser.in_eof = 0; \
	fawk_close_include(ctx, ctx->parser.isp); \
	if (ctx->parser.isp == &ctx->parser.include_stack[0]) { fawk_free(ctx, ctx->parser.buff); ctx->parser.buff = NULL; ctx->parser.alloced = ctx->parser.used = 0; return EOF; } \
	ctx->parser.isp--; \
	{ restart_stmt; }

#define lex_textblk_start(restart_stmt) \
	if ((nchr = getch(ctx)) != '[') { ungetch(ctx, nchr); return '['; } \
	ctx->parser.in_textblk = getch(ctx); ctx->parser.textblk_state = 0; restart_stmt;

/* textblk_state bitfield: 1=escape to expr, 2=need concat, 4 = last char was control char */
#define lex_textblk_exec(token_str, restart_stmt, nlchar)    if (ctx->parser.in_textblk) { \
		if (((chr = getch(ctx)) == ']') && (ctx->parser.textblk_state & 4)) { \
			if (getch(ctx) != ']') { LIBFAWK_ERROR(ctx, "expected a second ']' for closing text block", ctx->parser.isp->fn, ctx->parser.isp->line+1, ctx->parser.isp->col+1,  -1); } \
			ctx->parser.in_textblk = 0; restart_stmt; \
		} \
		ctx->parser.textblk_state &= ~4; \
		if (chr == ctx->parser.in_textblk) { ctx->parser.textblk_state |= 4; } \
		if (ctx->parser.textblk_state & 2) { ungetch(ctx, chr); ctx->parser.textblk_state &= ~2; return '@'; } \
		textblk_closechar:; if (ctx->parser.textblk_state & 1) { \
			if (chr == ctx->parser.in_textblk) { ctx->parser.textblk_state &= ~1; ctx->parser.textblk_state |= 2; ctx->parser.textblk_state |= 4; restart_stmt; } \
			ungetch(ctx, chr); \
		} else { \
			ctx->parser.textblk_state |= (1 | 2 | 4); \
			ungetch(ctx, chr); ctx->parser.used = 0; read_strlit(ctx, ctx->parser.in_textblk); \
			lval->str = fawk_strdup(ctx, ctx->parser.buff); return (lval->str == NULL) ? -1 : token_str; \
		} \
	} \
	do { chr = getch(ctx); } while((chr == ' ') || (chr == '\t') || (chr == '\r') || (chr == nlchar)); \
	if (chr == ctx->parser.in_textblk) { ctx->parser.textblk_state |= 4; goto textblk_closechar; }
